"""
This module's handler executor functions provide the interface to connect
`functions_framework` to the `ucam_faas` handler functions.

The handler executor functions are the base level handlers to layer
higher-level, (more specific) handlers on top of. This module's handlers support
reporting function results from ExecutionInfo objects. The results are made
available in 3 ways:

- As HTTP responses to the HTTP request that triggered the event handler
- As structured (JSON) log records
- As the original ExecutionInfo value for unit tests (tests need not
  mock or use a full functions_framework server to run a function to
  test it)
"""

from __future__ import annotations

import http.client
import logging
import time
import types
from abc import abstractmethod
from dataclasses import InitVar, dataclass, field
from enum import Enum
from typing import TYPE_CHECKING, Callable, Final, Generic, Mapping, Protocol

import flask
import flask.typing
import functions_framework
from cloudevents.abstract import CloudEvent
from flask import Response
from functions_framework.execution_id import EXECUTION_ID_REQUEST_HEADER
from pydantic_core import PydanticSerializationError
from structlog.typing import FilteringBoundLogger
from typing_extensions import ParamSpec, TypeVar
from werkzeug.exceptions import HTTPException

from ucam_faas._logging import flatten_json, get_structlog_logger
from ucam_faas.exceptions import ExecutionAbortedException
from ucam_faas.execution_info import (
    AbortedExecutionReason,
    ExecutionInfo,
    ExecutionStatus,
)


class CreateExecutionResultHttpResponseFn(Protocol):
    def __call__(self, execution_info: ExecutionInfo, /) -> Response | None:
        pass


@dataclass
class HttpResponseCreator(CreateExecutionResultHttpResponseFn):
    """
    Create minimalistic HTTP responses to communicate execution result.

    Functions are triggered by an HTTP request containing a `CloudEvent`. We
    need to send an HTTP response that communicates whether our function handled
    the event successfully, or failed in some way.

    It's expected that the HTTP response is only used to indicate that the
    function:

    - did not complete normally, and should be retried
    - was invoked with invalid/incorrect event data
    - completed successfully

    These use-cases should only require setting a status code. To help humans
    with debugging, we also send a short prose description of execution result,
    as a plain-text message in the response body.
    """

    @dataclass
    class ResponseFields:
        status_code: int
        message: str

    responses: Mapping[AbortedExecutionReason | str, ResponseFields]

    def get_unknown_reason_response(self, execution_info: ExecutionInfo) -> ResponseFields:
        return HttpResponseCreator.ResponseFields(
            status_code=http.client.INTERNAL_SERVER_ERROR,
            message="Failed to handle event for unknown reason",
        )

    def __call__(self, execution_info: ExecutionInfo) -> Response | None:
        if execution_info.execution.status is ExecutionStatus.COMPLETED:
            return None  # default successful response

        response_fields = self.responses.get(
            execution_info.execution.reason
        ) or self.get_unknown_reason_response(execution_info)

        return Response(
            response=response_fields.message,
            content_type="text/plain",
            status=response_fields.status_code,
        )


DEFAULT_HTTP_RESPONSE_CREATOR = HttpResponseCreator(
    responses={
        AbortedExecutionReason.FUNCTION_ABORTED: HttpResponseCreator.ResponseFields(
            status_code=http.client.INTERNAL_SERVER_ERROR,
            message="An error occurred while handling the CloudEvent",
        ),
        AbortedExecutionReason.CLOUD_EVENT_TYPE_INCORRECT: HttpResponseCreator.ResponseFields(
            status_code=http.client.BAD_REQUEST,
            message="The CloudEvent sent to the function was not a type expected by the function.",
        ),
        AbortedExecutionReason.SERVICE_DATA_INVALID: HttpResponseCreator.ResponseFields(
            status_code=http.client.BAD_REQUEST,
            message="The service-specific event metadata in the CloudEvent's data "
            "payload was not structured correctly for the CloudEvent type.",
        ),
        AbortedExecutionReason.USER_DATA_INVALID: HttpResponseCreator.ResponseFields(
            status_code=http.client.BAD_REQUEST,
            message="The application-specific, end-user-supplied data in the "
            "service-specific event's data payload was not valid.",
        ),
    }
)


class ExecutionResultLoggerFn(Protocol):
    def __call__(self, execution_info: ExecutionInfo, /) -> None:
        pass


class JsonRepresentation(Enum):
    NARROW_DEEP = "deep"
    """
    JSON objects and arrays can be nested arbitrarily.

    For example: `{"foo": {"bar": true}, "baz": false}`
    """
    WIDE_FLAT = "wide"
    """
    JSON is transformed into a single wide object containing all the values.

    Each value's property is its dot-separated path in the original JSON.

    For example: `{"foo.bar": true, "baz": false}`
    """


@dataclass
class StructlogJsonExecutionResultLogger(ExecutionResultLoggerFn):
    """
    Record function execution results as structured JSON log events with structlog.
    """

    json_representation: JsonRepresentation = JsonRepresentation.NARROW_DEEP
    logger: FilteringBoundLogger = field(default_factory=get_structlog_logger)

    EVENT_NAME_ABORTED: Final[str] = "ucam_faas_function_aborted"
    EVENT_NAME_COMPLETED: Final[str] = "ucam_faas_function_completed"

    def get_result_json(self, execution_info: ExecutionInfo) -> dict[str, object]:
        try:
            nested_result_json = execution_info.model_dump(mode="json")
        except PydanticSerializationError:
            self.logger.exception("execution_info_json_serialisation_failed")
            raise

        if self.json_representation is JsonRepresentation.NARROW_DEEP:
            return nested_result_json

        return flatten_json(nested_result_json)

    def send_result_json(
        self, execution_status: ExecutionStatus, result_json: dict[str, object]
    ) -> None:
        level, event_name = (
            (logging.ERROR, self.EVENT_NAME_ABORTED)
            if execution_status is ExecutionStatus.ABORTED
            else (logging.INFO, self.EVENT_NAME_COMPLETED)
        )

        # level and event last to prevent dynamic JSON properties overriding them
        self.logger.log(**result_json, level=level, event=event_name)

    def __call__(self, execution_info: ExecutionInfo) -> None:
        self.send_result_json(
            execution_info.execution.status, result_json=self.get_result_json(execution_info)
        )


DEFAULT_EXECUTION_RESULT_LOGGER = StructlogJsonExecutionResultLogger()


@dataclass
class HandlerExecutionOptions:
    log_execution_result: ExecutionResultLoggerFn | None = None
    make_http_response: CreateExecutionResultHttpResponseFn | None = None


_P = ParamSpec("_P")
_HttpResponseT = TypeVar("_HttpResponseT")


@dataclass
class FunctionsFrameworkHandlerExecutor(Generic[_P, _HttpResponseT]):
    """
    A generic `functions_framework` event handler that uses `ExecutionInfo` to
    report results.

    When called with an event, this calls its event-handling function, which
    returns an ExecutionInfo or raises an ExecutionAbortedError, both of which
    result in an ExecutionInfo value.

    The handler is responsible for logging the result of the execution,
    and creating an HTTP response in a manner compatible with being called as a
    functions_framework event handler.
    """

    log_execution_result: ExecutionResultLoggerFn
    create_http_response: CreateExecutionResultHttpResponseFn
    register: InitVar[bool]
    """Register this function with a `@functions_framework.*` handler decorator.

    If this is false, `functions_framework` will not be aware of the instance
    unless it's explicitly registered by the caller.
    """

    def __post_init__(self, register: bool) -> None:
        if register:
            # functions_framework decorators (like @cloud_event) register metadata
            # to make functions available as event handlers, but the decorator's
            # return value is not significant, as they just wrap the input with an
            # identity function. So it's fine to ignore the return value of the
            # decorator, and call us directly. The name of the function is used
            # to uniquely identify it and resolve it from a module.
            assert self.__name__ == self.execute.__name__
            self.register_with_functions_framework()  # nosemgrep: bandit.B101

    @abstractmethod
    def register_with_functions_framework(self) -> None:
        pass

    @abstractmethod
    def send_http_response(self, response: flask.Response | None) -> _HttpResponseT:
        pass

    if TYPE_CHECKING:

        @property
        @abstractmethod
        def execute(self) -> Callable[_P, ExecutionInfo]:
            ...

    def execute_for_result(self, *args: _P.args, **kwargs: _P.kwargs) -> ExecutionInfo:
        """
        Call the handler function and return its `ExecutionInfo`.

        The handler function is called in the same way as a real call, except no
        HTTP response is sent. Timing and result logging is enabled.
        """
        start_time_seconds = time.perf_counter()

        # We need to ensure we always catch and handle exceptions with
        # ExecutionAbortedException, re-throwing if required. Or maybe we do
        # this logging within the call().
        # Middle ground: catch within the call, but do side-effect logging and
        # HTTP response outside the call.
        # (This is now what we're dong.)

        execution_info: ExecutionInfo
        try:
            execution_info = self.execute(*args, **kwargs)
        except ExecutionAbortedException as e:
            execution_info = e.execution_info
        # The handler is expected to be responsible for catching other
        # exceptions and re-throwing them, wrapped in ExecutionAbortedException
        # in order to communicate an execution_info.

        execution_info.execution.duration_seconds = time.perf_counter() - start_time_seconds

        self.log_execution_result(execution_info)

        return execution_info

    def __call__(self, *args: _P.args, **kwargs: _P.kwargs) -> _HttpResponseT:
        execution_info = self.execute_for_result(*args, **kwargs)
        response = self.create_http_response(execution_info)
        return self.send_http_response(response)

    # @functions_framework.cloud_event() decorator registers metadata using the
    # __name__ of the decorated function, so we need to preserve the name of the
    # message_handler, so that our user can use the name of their function as
    # its identifier when configuring/deploying their cloud function.
    @property
    def __name__(self) -> str:
        return self.execute.__name__

    @property  # type: ignore[misc]
    def __class__(self) -> type[types.FunctionType]:  # type: ignore[override]
        # functions_framework is over-zealous with runtime type checking, it
        # requires that registered functions are instances of types.FunctionType,
        # not just that they are `callable()`. To work around this, we falsely
        # report our executor to be a FunctionType subclass by reporting our
        # instances' __class__ is types.FunctionType.
        return types.FunctionType


_CloudEventHandlerT = TypeVar("_CloudEventHandlerT", bound=Callable[[CloudEvent], ExecutionInfo])
_CloudEventHandlerT_co = TypeVar(
    "_CloudEventHandlerT_co", bound=Callable[[CloudEvent], ExecutionInfo], covariant=True
)


class CloudEventFunctionsFrameworkHandlerExecutorFn(Protocol[_CloudEventHandlerT_co]):
    def __call__(self, cloud_event: CloudEvent) -> None:
        pass

    if TYPE_CHECKING:

        @property
        def __name__(self) -> str:
            pass

        @property
        def execute(self) -> _CloudEventHandlerT_co:
            """
            Call the handler function and return its `ExecutionInfo`.

            The normal post-handler steps of logging the result and sending an
            HTTP response are not performed.
            """

        def execute_for_result(self, cloud_event: CloudEvent) -> ExecutionInfo:
            """
            Call the handler function and return its `ExecutionInfo`.

            The handler function is called in the same way as a real call, except no
            HTTP response is sent. Timing and result logging is enabled.
            """


def cloud_event_handler_executor(
    *,
    log_execution_result: ExecutionResultLoggerFn,
    create_http_response: CreateExecutionResultHttpResponseFn,
) -> Callable[
    [_CloudEventHandlerT], CloudEventFunctionsFrameworkHandlerExecutorFn[_CloudEventHandlerT]
]:
    def cloud_event_handler_executor_decorator(
        handler: _CloudEventHandlerT, /
    ) -> CloudEventFunctionsFrameworkHandlerExecutorFn[_CloudEventHandlerT]:
        decorated: CloudEventFunctionsFrameworkHandlerExecutorFn[
            _CloudEventHandlerT
        ] = CloudEventFunctionsFrameworkHandlerExecutor(
            log_execution_result=log_execution_result,
            create_http_response=create_http_response,
            register=True,
            execute=handler,
        )
        return decorated

    return cloud_event_handler_executor_decorator


@dataclass
class CloudEventFunctionsFrameworkHandlerExecutor(
    FunctionsFrameworkHandlerExecutor[[CloudEvent], None],
    CloudEventFunctionsFrameworkHandlerExecutorFn[_CloudEventHandlerT],
    Generic[_CloudEventHandlerT],
):
    """
    A generic `@functions_framework.cloud_event` event handler that uses
    `ExecutionInfo` to report results.
    """

    execute: _CloudEventHandlerT

    def register_with_functions_framework(self) -> None:
        functions_framework.cloud_event(self)

    def send_http_response(self, response: flask.Response | None) -> None:
        if response is not None:
            raise HTTPException(response=response)


_HttpHandlerT = TypeVar("_HttpHandlerT", bound=Callable[[flask.Request], ExecutionInfo])
_HttpHandlerT_co = TypeVar(
    "_HttpHandlerT_co", bound=Callable[[flask.Request], ExecutionInfo], covariant=True
)


class HttpFunctionsFrameworkHandlerExecutorFn(Protocol[_HttpHandlerT_co]):
    def __call__(self, request: flask.Request) -> flask.typing.ResponseReturnValue:
        pass

    if TYPE_CHECKING:

        @property
        def __name__(self) -> str:
            pass

        @property
        def execute(self) -> _HttpHandlerT_co:
            """
            Call the handler function and return its `ExecutionInfo`.

            The normal post-handler steps of logging the result and sending an
            HTTP response are not performed.
            """

        def execute_for_result(self, request: flask.Request) -> ExecutionInfo:
            """
            Call the handler function and return its `ExecutionInfo`.

            The handler function is called in the same way as a real call, except no
            HTTP response is sent. Timing and result logging is enabled.
            """


@dataclass
class HttpFunctionsFrameworkHandlerExecutor(
    FunctionsFrameworkHandlerExecutor[[flask.Request], flask.typing.ResponseReturnValue],
    HttpFunctionsFrameworkHandlerExecutorFn[_HttpHandlerT],
    Generic[_HttpHandlerT],
):
    """
    A generic `@functions_framework.http` event handler that uses `ExecutionInfo`
    to report results.

    The handler is triggered by an arbitrary HTTP request, rather than a specific
    CloudEvent payload. Unlike a conventional HTTP API server, the handler is
    expected to be executed primarily for its side-effect (handling an event),
    and its HTTP response is not expected to represent an HTTP resource, rather
    it should only indicate success or failure in handling the request.

    A typical use-case would be to receive an HTTP web hook notification request,
    in which the web hook sender may retry failed requests, but does not expect
    to receive structured data in response.

    Note
    ----
    This exists as a proof-of-concept to ensure the
    `FunctionsFrameworkHandlerExecutor` is able to support use cases other than
    `functions_framework.cloud_event()`. There's no use-friendly decorator to
    register HTTP event handlers at the moment.
    """

    execute: _HttpHandlerT

    def register_with_functions_framework(self) -> None:
        functions_framework.http(self)

    def send_http_response(
        self, response: flask.Response | None
    ) -> flask.typing.ResponseReturnValue:
        if response is not None:
            return response
        return flask.Response()


def get_function_execution_id() -> str | None:
    """
    Get the unique identifier for the currently-executing GCP Cloud Run function.

    Notes
    -----
    See `functions_framework.execution_id` module. `functions_framework` always
    generates a value if one isn't set, but it should be provided by Cloud Run
    in Cloud Run deployments.
    """
    if flask.has_request_context():
        return flask.request.headers.get(EXECUTION_ID_REQUEST_HEADER) or None
    return None



import re
from datetime import datetime
from typing import Any

import pytest
from pydantic import TypeAdapter, ValidationError

from ucam_faas.gcp_pubsub import Base64JsonString, MessagePublishedData, PubsubMessage

datetime_type = TypeAdapter(datetime)


def test_Base64JsonString() -> None:
    base64_json_string = TypeAdapter(Base64JsonString)

    # JSON representation is a base64-encoded string
    assert base64_json_string.validate_json('"bWVzc2FnZTE="') == b"message1"
    assert base64_json_string.dump_json(b"message1") == b'"bWVzc2FnZTE="'
    assert base64_json_string.dump_python(b"message1", mode="json") == "bWVzc2FnZTE="

    # The Python representation is bytes
    assert base64_json_string.validate_python(b"\x00foobar\xff") == b"\x00foobar\xff"
    assert base64_json_string.dump_python(b"message1") == b"message1"

    # A Python str is also accepted, and decoded as bytes
    assert base64_json_string.validate_python("bWVzc2FnZTE=") == b"message1"

    # Invalid base64 values are reported as pydantic ValidationError
    with pytest.raises(ValidationError) as exc_info:
        base64_json_string.validate_python("00000")  # invalid base64

    assert "Invalid base64-encoded string: " in str(exc_info.value)
    assert exc_info.value.errors()[0]["type"] == "base64_decode"


tuple[dict[str, object], MessagePublishedData]
valid_pubsub_message_published_event_data_examples = pytest.mark.parametrize(
    "message_json,message_parsed",
    [
        pytest.param(
            {
                "message": {
                    "@type": "type.googleapis.com/google.pubsub.v1.PubsubMessage",
                    "data": "bWVzc2FnZTE=",
                    "attributes": {"my_attr_1": "foo", "my_attr_2": "bar"},
                    "messageId": "1",
                    "publishTime": "2025-03-17T15:07:22.650Z",
                }
            },
            MessagePublishedData(
                message=PubsubMessage(
                    data=b"message1",
                    attributes={"my_attr_1": "foo", "my_attr_2": "bar"},
                    message_id="1",
                    publish_time=datetime_type.validate_python("2025-03-17T15:07:22.650Z"),
                ),
            ),
            id="with-attrs",
        ),
        pytest.param(
            {
                "message": {
                    "@type": "type.googleapis.com/google.pubsub.v1.PubsubMessage",
                    "data": "AAECAwQFBgcICQoLDA0ODxAREhMUFRYXGBkaGxwdHh8gISIjJC"
                    "UmJygpKissLS4vMDEyMzQ1Njc4OTo7PD0+P0BBQkNERUZHSElKS0xNTk9Q"
                    "UVJTVFVWV1hZWltcXV5fYGFiY2RlZmdoaWprbG1ub3BxcnN0dXZ3eHl6e3"
                    "x9fn+AgYKDhIWGh4iJiouMjY6PkJGSk5SVlpeYmZqbnJ2en6ChoqOkpaan"
                    "qKmqq6ytrq+wsbKztLW2t7i5uru8vb6/wMHCw8TFxsfIycrLzM3Oz9DR0t"
                    "PU1dbX2Nna29zd3t/g4eLj5OXm5+jp6uvs7e7v8PHy8/T19vf4+fr7/P3+",
                    "attributes": {},
                    "messageId": "2",
                    "publishTime": "2025-03-17T15:10:35.306Z",
                }
            },
            MessagePublishedData(
                message=PubsubMessage(
                    data=bytes(range(255)),
                    attributes={},
                    message_id="2",
                    publish_time=datetime_type.validate_python("2025-03-17T15:10:35.306Z"),
                ),
            ),
            id="without-attrs",
        ),
    ],
)


@valid_pubsub_message_published_event_data_examples
def test_MessagePublishedData_round_trip(
    message_json: dict[str, object], message_parsed: MessagePublishedData
) -> None:
    parsed = MessagePublishedData.model_validate(message_json)
    assert parsed == message_parsed

    dumped = parsed.model_dump(mode="json", by_alias=True)
    # Pydantic uses more digits to represent fractional sections than we get in
    # real payload data.
    dumped["message"]["publishTime"] = _normalise_iso_datetime(dumped["message"]["publishTime"])
    assert message_json == dumped


@valid_pubsub_message_published_event_data_examples
def test_PubsubMessage_timestamps_are_tz_aware(
    message_json: dict[str, object], message_parsed: MessagePublishedData
) -> None:
    instance = MessagePublishedData.model_validate(message_json)
    assert instance.message.publish_time.tzinfo is not None


@valid_pubsub_message_published_event_data_examples
def test_PubsubMessage_validation_rejects_non_tz_aware_timestamps(
    message_json: dict[str, Any], message_parsed: MessagePublishedData
) -> None:
    message_json["message"]["publishTime"] = message_json["message"]["publishTime"].rstrip("Z")
    with pytest.raises(ValidationError, match=r"Input should have timezone info"):
        MessagePublishedData.model_validate(message_json)

    with pytest.raises(ValidationError, match=r"Input should have timezone info"):
        PubsubMessage(**message_parsed.model_dump(), publish_time=datetime.now(tz=None))


@valid_pubsub_message_published_event_data_examples
def test_PubsubMessage_message_id_cannot_be_empty(
    message_json: dict[str, Any], message_parsed: MessagePublishedData
) -> None:
    message_json["message"]["messageId"] = ""
    with pytest.raises(ValidationError, match=r"should have at least 1 character"):
        MessagePublishedData.model_validate(message_json)

    with pytest.raises(ValidationError, match=r"should have at least 1 character"):
        PubsubMessage(**message_parsed.model_dump(), message_id="")


@valid_pubsub_message_published_event_data_examples
def test_PubsubMessage_type_must_be_correct(
    message_json: dict[str, Any], message_parsed: MessagePublishedData
) -> None:
    message_json["message"]["@type"] = "foobar"
    with pytest.raises(ValidationError, match=r"Input should be.+PubsubMessage"):
        MessagePublishedData.model_validate(message_json)

    with pytest.raises(ValidationError, match=r"Input should be.+PubsubMessage"):
        PubsubMessage(**message_parsed.model_dump(), type="foobar")


def _normalise_iso_datetime(iso_datetime: str) -> str:
    """Truncate fractional seconds of an ISO datetime string to 3 digits.

    Examples
    --------
    >>> _normalise_iso_datetime('2025-03-17T15:10:35.306000Z')
    '2025-03-17T15:10:35.306Z'
    >>> _normalise_iso_datetime('asdfsdfds')
    'asdfsdfds'
    """
    # limit fractional seconds to 1ms resolution (3 digits)
    return re.sub(r"\.([0-9]+)[zZ]$", lambda m: f".{m.group(1)[:3]}Z", iso_datetime)



import pytest
from cloudevents.abstract import CloudEvent

from ucam_faas.tests.fixtures import make_cloud_event

# Register the ucam_faas testing module as a pytest plugin - as it provides
# fixtures that we can make use of when testing the functions.
pytest_plugins = ["ucam_faas.testing"]


@pytest.fixture()
def valid_cloud_event() -> CloudEvent:
    return make_cloud_event(type="ucam_faas_event", source="ucam_faas", data={"foo": "bar"})



from __future__ import annotations

import gzip
import json
import re
from typing import Callable, Generic, Literal, TypeVar, cast

import pytest
from google.protobuf.message import DecodeError
from pydantic import BaseModel, Field
from typing_extensions import Never, assert_type

from ucam_faas.cloud_events import CloudEventAttributes, CloudEventHandlerExecutionInfo
from ucam_faas.contexts import contextual_default_factory
from ucam_faas.exceptions import ExecutionAbortedException
from ucam_faas.execution_info import (
    AbortedExecutionReason,
    AbortedExecutionResult,
    ExecutionStatus,
)
from ucam_faas.messages import (
    MESSAGE_CONTEXT,
    MessageHandlerExecutionInfo,
    message_handler,
)
from ucam_faas.tests.fixtures import (
    make_cloud_event,
    make_pubsub_cloud_event,
    make_pubsub_message,
)
from ucam_faas.tests.protobuf import example_pb2

_T = TypeVar("_T")


class ExampleMessage(BaseModel):
    colour: Literal["red", "green", "blue"]
    size: int


class EchoExecutionInfo(MessageHandlerExecutionInfo, Generic[_T]):
    message: _T


def echo_message_handler(message: _T) -> EchoExecutionInfo[_T]:
    return EchoExecutionInfo(message=message)


def gzip_json_message_parser(raw_message: bytes) -> dict[str, object]:
    json_value = json.loads(gzip.decompress(raw_message))
    assert isinstance(json_value, dict)
    return json_value


def aborting_message_parser(raw_message: bytes) -> Never:
    raise ExecutionAbortedException(
        "aborted due to testing",
        execution_info=MessageHandlerExecutionInfo(
            execution=AbortedExecutionResult(
                status=ExecutionStatus.ABORTED, reason="custom-reason"
            )
        ),
    )


def value_error_message_parser(raw_message: bytes) -> Never:
    raise ValueError("no message data values are valid")


def test_message_handler__with_no_message_parser__provides_bytes_to_message_handler() -> None:
    @message_handler(message_type=None)
    def demo_handler(message: bytes) -> EchoExecutionInfo[bytes]:
        return EchoExecutionInfo(message=message)

    pubsub_cloud_event = make_pubsub_cloud_event(make_pubsub_message(data=b"foo"))
    execution_info = demo_handler.execute(pubsub_cloud_event)

    assert execution_info.execution.status == ExecutionStatus.COMPLETED
    assert execution_info.message == b"foo"


def test_message_handler__with_pydantic_message_parser__provides_model_instance_to_message_handler() -> (  # noqa: E501
    None
):
    @message_handler(message_type=ExampleMessage)
    def demo_handler(message: ExampleMessage) -> EchoExecutionInfo[ExampleMessage]:
        return EchoExecutionInfo(message=message)

    model_msg = ExampleMessage(colour="red", size=42)
    execution_info = demo_handler.execute(
        make_pubsub_cloud_event(make_pubsub_message(data=model_msg))
    )

    assert execution_info.execution.status == ExecutionStatus.COMPLETED
    assert execution_info.message == model_msg


def test_message_handler__with_pydantic_message_parser__rejects_message_with_invalid_data() -> (
    None
):
    @message_handler(message_type=ExampleMessage)
    def demo_handler(message: ExampleMessage) -> EchoExecutionInfo[ExampleMessage]:
        return EchoExecutionInfo(message=message)

    with pytest.raises(
        ExecutionAbortedException, match=r"Input should be 'red', 'green' or 'blue'"
    ) as exc_info:
        demo_handler.execute(
            make_pubsub_cloud_event(
                # Invalid JSON data for ExampleMessage
                make_pubsub_message(data={"colour": "foo", "size": 42})
            )
        )

    execution_info = exc_info.value.execution_info
    assert execution_info.execution.status == ExecutionStatus.ABORTED
    assert execution_info.execution.reason == AbortedExecutionReason.USER_DATA_INVALID


def test_message_handler__with_protobuf_message_parser__provides_message_instance_to_message_handler() -> (  # noqa: E501
    None
):
    @message_handler(message_type=example_pb2.ExampleRecord)
    def demo_handler(
        message: example_pb2.ExampleRecord,
    ) -> EchoExecutionInfo[example_pb2.ExampleRecord]:
        return EchoExecutionInfo(message=message)

    protobuf_msg = example_pb2.ExampleRecord(query="foo", page_number=3, results_per_page=10)
    execution_info = demo_handler.execute(
        make_pubsub_cloud_event(make_pubsub_message(data=protobuf_msg))
    )

    assert execution_info.execution.status == ExecutionStatus.COMPLETED
    assert execution_info.message == protobuf_msg


def test_message_handler__with_protobuf_message_parser__rejects_message_with_invalid_data() -> (  # noqa: E501
    None
):
    @message_handler(message_type=example_pb2.ExampleRecord)
    def demo_handler(
        message: example_pb2.ExampleRecord,
    ) -> EchoExecutionInfo[example_pb2.ExampleRecord]:
        return EchoExecutionInfo(message=message)

    with pytest.raises(
        ExecutionAbortedException, match=r"message_parser failed to parse message"
    ) as exc_info:
        demo_handler.execute(
            # 0xFF is not a valid serialised example_pb2.ExampleRecord
            make_pubsub_cloud_event(make_pubsub_message(data=b"\xff"))
        )
    execution_info = exc_info.value.execution_info
    assert execution_info.execution.status == ExecutionStatus.ABORTED
    assert isinstance(execution_info.execution.exception, DecodeError)
    assert execution_info.execution.reason == AbortedExecutionReason.USER_DATA_INVALID


def test_message_handler__with_function_message_parser__provides_message_instance_to_message_handler() -> (  # noqa: E501
    None
):
    @message_handler(message_type=gzip_json_message_parser)
    def demo_handler(message: dict[str, object]) -> EchoExecutionInfo[dict[str, object]]:
        return EchoExecutionInfo(message=message)

    custom_msg = {"hello": "world"}
    encoded_custom_msg = gzip.compress(json.dumps(custom_msg).encode())
    execution_info = demo_handler.execute(
        make_pubsub_cloud_event(make_pubsub_message(data=encoded_custom_msg))
    )

    assert execution_info.execution.status == ExecutionStatus.COMPLETED
    assert execution_info.message == custom_msg


def test_message_handler__with_function_message_parser__rejects_message_with_invalid_data() -> (  # noqa: E501
    None
):
    @message_handler(message_type=gzip_json_message_parser)
    def demo_handler(message: dict[str, object]) -> EchoExecutionInfo[dict[str, object]]:
        return EchoExecutionInfo(message=message)

    with pytest.raises(
        ExecutionAbortedException, match=re.escape(r"Not a gzipped file (b'\xff')")
    ) as exc_info:
        demo_handler.execute(
            # 0xFF is not valid gzip stream
            make_pubsub_cloud_event(make_pubsub_message(data=b"\xff"))
        )
    execution_info = exc_info.value.execution_info
    assert execution_info.execution.status == ExecutionStatus.ABORTED
    assert execution_info.execution.reason == AbortedExecutionReason.USER_DATA_INVALID


def test_message_handler__with_function_message_parser__honours_AbortedExecutionResult_thrown_by_parser() -> (  # noqa: E501
    None
):
    @message_handler(message_type=aborting_message_parser)
    def demo_handler(message: object) -> EchoExecutionInfo[Never]:
        assert False, "not reachable due to parser aborting"

    with pytest.raises(ExecutionAbortedException, match=r"aborted due to testing") as exc_info:
        demo_handler.execute(make_pubsub_cloud_event(make_pubsub_message(data=b"")))
    execution_info = exc_info.value.execution_info
    assert execution_info.execution.status == ExecutionStatus.ABORTED
    assert execution_info.execution.reason == "custom-reason"


def test_message_handler__with_invalid_message_parser__throws_type_error_on_creation() -> None:
    with pytest.raises(
        TypeError,
        match=r"message_parser must be "
        r"a Pydantic model class, "
        r"a Protobuf Message class, "
        r"a function "
        r"or None, got: <object object at .+>",
    ):

        @message_handler(message_type=cast(None, object()))
        def demo_handler(message: bytes) -> Never:
            assert False, "Never called due to invalid message_type"


def test_message_handler__can_return_none_to_use_default_execution_info_result() -> None:
    received_message: bytes | None = None

    @message_handler(message_type=None)
    def handle_bytes(message: bytes) -> None:
        nonlocal received_message
        received_message = message

    pubsub_cloud_event = make_pubsub_cloud_event(make_pubsub_message(data=b"foo"))
    execution_info = handle_bytes.execute(pubsub_cloud_event)

    assert_type(execution_info, MessageHandlerExecutionInfo)
    assert isinstance(execution_info, MessageHandlerExecutionInfo)

    assert execution_info.execution.status is ExecutionStatus.COMPLETED
    assert received_message == b"foo"


def test_message_handler__returns_specified_default_execution_info_type() -> None:
    """
    The handler returns no ExecutionInfo itself, so a default ExecutionInfo gets
    created after it returns.

    The default ExecutionInfo type is overridden to be MyExecutionInfo, so it
    gets created, and picks up an attribute from the parsed Pubsub message to
    record as part of the result.
    """

    class MyExecutionInfo(MessageHandlerExecutionInfo):
        foo: str = Field(
            default_factory=contextual_default_factory(
                MESSAGE_CONTEXT, fn=lambda msg: msg.pubsub_message.attributes["foo"]
            )
        )

    @message_handler(message_type=None, create_default_info=MyExecutionInfo.from_context)
    def demo_handler(message: bytes) -> None:
        pass

    pubsub_cloud_event = make_pubsub_cloud_event(
        make_pubsub_message(data=b"foo", attributes={"foo": "foobar!"})
    )
    execution_info = demo_handler.execute(pubsub_cloud_event)

    assert_type(execution_info, MyExecutionInfo)
    assert isinstance(execution_info, MyExecutionInfo)

    assert execution_info.execution.status is ExecutionStatus.COMPLETED
    assert execution_info.foo == "foobar!"


@pytest.mark.parametrize("message_type", [aborting_message_parser, value_error_message_parser])
def test_message_handler__provides_input_message_in_execution_info__when_input_message_is_rejected(
    message_type: Callable[[bytes], Never],
) -> None:
    @message_handler(message_type=message_type)
    def demo_handler(message: bytes) -> EchoExecutionInfo[bytes]:
        assert False, "not reachable due to parser aborting"

    pubsub_message = make_pubsub_message(data=b"foo")
    pubsub_cloud_event = make_pubsub_cloud_event(pubsub_message)

    with pytest.raises(ExecutionAbortedException) as exc_info:
        demo_handler.execute(pubsub_cloud_event)

    execution_info = exc_info.value.execution_info
    assert isinstance(execution_info, MessageHandlerExecutionInfo)
    assert execution_info.execution.status == ExecutionStatus.ABORTED

    assert execution_info.cloudevents == CloudEventAttributes.from_cloud_event(pubsub_cloud_event)
    assert execution_info.cloud_event_handler.input_event_accepted is True

    assert execution_info.message_handler.input_message_accepted is False
    assert execution_info.pubsub_message == pubsub_message


def test_message_handler__provides_input_message_in_execution_info__when_input_message_is_accepted() -> (  # noqa: E501
    None
):
    @message_handler(message_type=None)
    def demo_handler(message: bytes) -> None:
        return None

    pubsub_message = make_pubsub_message(data=b"foo")
    pubsub_cloud_event = make_pubsub_cloud_event(pubsub_message)

    execution_info = demo_handler.execute(pubsub_cloud_event)

    assert execution_info.execution.status == ExecutionStatus.COMPLETED

    assert execution_info.cloudevents == CloudEventAttributes.from_cloud_event(pubsub_cloud_event)
    assert execution_info.cloud_event_handler.input_event_accepted is True

    assert execution_info.message_handler.input_message_accepted is True
    assert execution_info.pubsub_message == pubsub_message


def test_message_handler__provides_cloud_event_only__when_input_cloud_event_is_not_pubsub_message_event() -> (  # noqa: E501
    None
):
    @message_handler(message_type=None)
    def demo_handler(message: bytes) -> None:
        return None

    wrong_cloud_event = make_cloud_event(
        type="com.example.sampletype1", source="https://example.com/event-producer"
    )

    with pytest.raises(
        ExecutionAbortedException, match="demo_handler input parser rejected input"
    ) as exc_info:
        demo_handler.execute(wrong_cloud_event)

    execution_info = exc_info.value.execution_info
    assert isinstance(execution_info, CloudEventHandlerExecutionInfo)
    assert not isinstance(execution_info, MessageHandlerExecutionInfo)
    assert execution_info.execution.status == ExecutionStatus.ABORTED

    assert execution_info.cloudevents == CloudEventAttributes.from_cloud_event(wrong_cloud_event)
    assert execution_info.cloud_event_handler.input_event_accepted is False



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Literal, Union

from pydantic import BaseModel, ConfigDict, Field
from typing_extensions import Self, TypeAlias

from ucam_faas.pydantic import SerialisableException


class ExecutionStatus(Enum):
    COMPLETED = "COMPLETED"
    ABORTED = "ABORTED"


class AbortedExecutionReason(Enum):
    FUNCTION_ABORTED = "ucam_faas.function-aborted"
    """The handler function threw an Exception, or returned an aborted ExecutionInfo."""
    CLOUD_EVENT_TYPE_INCORRECT = "ucam_faas.cloud-event-type-incorrect"
    """The CloudEvent sent to the function was not a type expected by the function."""
    SERVICE_DATA_INVALID = "ucam_faas.service-data-invalid"
    """The service-specific event metadata in the CloudEvent's data payload was not
    structured correctly for the CloudEvent type.
    """
    USER_DATA_INVALID = "ucam_faas.user-data-invalid"
    """The application-specific, end-user-supplied data in the service-specific
    event's data payload was rejected by the `message_validator`.
    """


@dataclass
class CompletedExecutionResult:
    """The function executed and returned normally."""

    status: Literal[ExecutionStatus.COMPLETED] = ExecutionStatus.COMPLETED
    duration_seconds: float | None = None


@dataclass
class AbortedExecutionResult:
    """The function was aborted before completing its work."""

    reason: AbortedExecutionReason | str
    status: Literal[ExecutionStatus.ABORTED] = ExecutionStatus.ABORTED
    exception: SerialisableException | None = None
    """The Exception that aborted the function execution.

    Can be None if the function explicitly returned an aborted ExecutionInfo
    value without raising, in which case the value should contain properties to
    describe what happened.
    """
    duration_seconds: float | None = None


ExecutionResult: TypeAlias = Union[CompletedExecutionResult, AbortedExecutionResult]


class ExecutionInfo(BaseModel):
    """Data to inform observers of the outcome of a function execution.

    - Handlers can return subclasses of this type, with appropriate fields to
      describe the result or side-effects of their function.
    - `ucam_faas` serializes values as JSON and logs them to provide operational
      observability of the behaviour of deployed functions.
    - Unit tests can access the un-serialized values to make assertions about
      the behaviour of a handler, given an input event object.
      - Use the `.execute(cloud_event)` method on the function decorated with
        `@message_handler(...)` to execute the function and return your info.
    - By providing plenty of info on your handler's outcome, you can improve
      both runtime observability and development testability in one go.
    """

    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)

    execution: ExecutionResult = Field(default_factory=CompletedExecutionResult)

    @classmethod
    def from_context(cls, *, execution: ExecutionResult) -> Self:
        """
        Create an instance of this ExecutionInfo type from the executing function's context.

        The ExecutionInfo class does not have any contextual fields, but
        subclasses should use `Field(default_factory=contextual_default_factory(...))`
        to automatically populate fields from `ContextVar`s they are expected to
        be created under.

        Notes
        -----
        In theory this shouldn't need to exist, as a reference to the class
        itself acts as a function that creates an instance with an `execution`,
        but Python type checkers seem to not match overloaded functions when a
        model class constructor is used as a function reference. Perhaps because
        of the Pydantic mypy plugin generating the constructor signature.
        """
        return cls(execution=execution)



from typing import ClassVar as _ClassVar
from typing import Optional as _Optional

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message

DESCRIPTOR: _descriptor.FileDescriptor

class ExampleRecord(_message.Message):
    __slots__ = ["page_number", "query", "results_per_page"]
    PAGE_NUMBER_FIELD_NUMBER: _ClassVar[int]
    QUERY_FIELD_NUMBER: _ClassVar[int]
    RESULTS_PER_PAGE_FIELD_NUMBER: _ClassVar[int]
    page_number: int
    query: str
    results_per_page: int
    def __init__(
        self,
        query: _Optional[str] = ...,
        page_number: _Optional[int] = ...,
        results_per_page: _Optional[int] = ...,
    ) -> None: ...



from __future__ import annotations

import re
from contextvars import ContextVar
from dataclasses import dataclass
from typing import Generic, Literal, TypeVar

import pytest
from pydantic import BaseModel, Field
from typing_extensions import Never, assert_type

from ucam_faas.contexts import (
    ActivateContextVarFn,
    contextual_default_factory,
    contextvar_activator,
)
from ucam_faas.contextual_handlers import ContextualHandlerFn, contextual_handler
from ucam_faas.exceptions import ExecutionAbortedException
from ucam_faas.execution_info import (
    AbortedExecutionReason,
    AbortedExecutionResult,
    ExecutionInfo,
    ExecutionResult,
    ExecutionStatus,
)

_T = TypeVar("_T")


class ExampleEvent(BaseModel):
    """A fictional event type for demonstration purposes."""

    version: Literal["1.0", "1.1"]
    id: str
    metadata: dict[str, str]
    message: str


@dataclass
class ExampleContext:
    """Metadata available implicitly to handler functions."""

    event: ExampleEvent
    weather_today: Literal["wet", "dry"]


EXAMPLE_CONTEXT = ContextVar[ExampleContext]("EXAMPLE_CONTEXT")

activate_example_context = contextvar_activator(EXAMPLE_CONTEXT)


class ExampleExecutionInfo(ExecutionInfo, Generic[_T]):
    data: _T

    weather_today: Literal["wet", "dry"] = Field(
        default_factory=contextual_default_factory(
            EXAMPLE_CONTEXT, fn=lambda ctx: ctx.weather_today
        )
    )


@dataclass
class ExampleEventContextParser:
    weather_today: Literal["wet", "dry"]

    def __call__(
        self, raw_json_event: bytes
    ) -> tuple[ExampleEvent, ExampleContext] | ExecutionInfo:
        try:
            event = ExampleEvent.model_validate_json(raw_json_event)
            if event.version != "1.0":
                raise ValueError(
                    f"Only event version 1.0 is not supported: version={event.version!r}"
                )
        except ValueError as e:
            result = AbortedExecutionResult(
                reason=AbortedExecutionReason.SERVICE_DATA_INVALID,
                exception=e,
            )
            return ExecutionInfo(execution=result)

        return (event, ExampleContext(event=event, weather_today=self.weather_today))


example_event_parser = ExampleEventContextParser(weather_today="dry")


@pytest.fixture
def example_event() -> ExampleEvent:
    return ExampleEvent(version="1.0", id="ab12", metadata={"location": "middle"}, message="Hi")


@pytest.fixture
def example_event_json(example_event: ExampleEvent) -> bytes:
    return example_event.model_dump_json().encode()


def test_contextual_handler__may_return_none(
    example_event: ExampleEvent, example_event_json: bytes
) -> None:
    @contextual_handler(
        parser=example_event_parser,
        context=activate_example_context,
        default_info_factory=ExecutionInfo.from_context,
    )
    def handle_event(event: ExampleEvent) -> None:
        assert event == example_event
        return None

    assert_type(
        handle_event,
        ContextualHandlerFn[bytes, ExampleEvent, Never, ExecutionInfo, None],
    )

    execution_info = handle_event(example_event_json)

    assert_type(execution_info, ExecutionInfo)
    assert execution_info.execution.status is ExecutionStatus.COMPLETED


def test_contextual_handler__returns_specified_default_ExecutionInfo_type_when_handler_returns_none(  # noqa: E501
    example_event: ExampleEvent, example_event_json: bytes
) -> None:
    def create_default_execution_info(
        *, execution: ExecutionResult
    ) -> ExampleExecutionInfo[Literal["foo"]]:
        return ExampleExecutionInfo[Literal["foo"]](data="foo", execution=execution)

    @contextual_handler(
        parser=example_event_parser,
        context=activate_example_context,
        default_info_factory=create_default_execution_info,
    )
    def handle_event(event: ExampleEvent) -> None:
        assert event == example_event
        return None

    assert_type(
        handle_event,
        ContextualHandlerFn[
            bytes, ExampleEvent, Never, ExampleExecutionInfo[Literal["foo"]], None
        ],
    )

    execution_info = handle_event(example_event_json)

    assert_type(execution_info, ExampleExecutionInfo[Literal["foo"]])
    assert execution_info.execution.status is ExecutionStatus.COMPLETED
    assert execution_info.data == "foo"
    assert execution_info.weather_today == example_event_parser.weather_today


def test_contextual_handler__returns_only_handler_type_when_handler_does_not_return_none(
    example_event: ExampleEvent, example_event_json: bytes
) -> None:
    @contextual_handler(
        parser=example_event_parser,
        context=activate_example_context,
        default_info_factory=ExecutionInfo.from_context,
    )
    def handle_event(event: ExampleEvent) -> ExampleExecutionInfo[str]:
        assert event == example_event
        return ExampleExecutionInfo(data=event.message)

    assert_type(
        handle_event,
        ContextualHandlerFn[bytes, ExampleEvent, ExampleExecutionInfo[str], Never, Never],
    )

    execution_info = handle_event(example_event_json)

    assert_type(execution_info, ExampleExecutionInfo[str])
    assert execution_info.execution.status == ExecutionStatus.COMPLETED
    assert execution_info.data == example_event.message
    assert execution_info.weather_today == example_event_parser.weather_today


@pytest.mark.parametrize("weather", ["wet", "dry"])
@pytest.mark.parametrize(
    "context",
    [
        pytest.param(EXAMPLE_CONTEXT, id="context-var"),
        pytest.param(activate_example_context, id="context-activator"),
    ],
)
def test_contextual_handler__provides_context_metadata_from_parser_in_context(
    example_event: ExampleEvent,
    example_event_json: bytes,
    weather: Literal["wet", "dry"],
    context: ContextVar[ExampleContext] | ActivateContextVarFn[ExampleContext],
) -> None:
    @contextual_handler(
        # the parser provides the value of EXAMPLE_CONTEXT
        parser=ExampleEventContextParser(weather_today=weather),
        # context can be a ContextVar or an contextvar_activator function
        context=context,
        default_info_factory=ExecutionInfo.from_context,
    )
    def handle_event(event: ExampleEvent) -> ExampleExecutionInfo[str]:
        assert event == example_event
        example_context = EXAMPLE_CONTEXT.get()
        assert example_context.event is event
        return ExampleExecutionInfo(data=example_context.weather_today)

    assert_type(
        handle_event,
        ContextualHandlerFn[bytes, ExampleEvent, ExampleExecutionInfo[str], Never, Never],
    )

    execution_info = handle_event(example_event_json)

    assert execution_info.data == weather
    assert execution_info.weather_today == weather


def test_contextual_handler__parser_can_return_ExecutionInfo_to_abort(
    example_event: ExampleEvent,
) -> None:
    # The parser returns if the version is not 1.0
    example_event.version = "1.1"
    unsupported_version_example_event_json = example_event.model_dump_json().encode()
    failed_parse_result = example_event_parser(unsupported_version_example_event_json)

    assert isinstance(failed_parse_result, ExecutionInfo)
    assert failed_parse_result.execution.status is ExecutionStatus.ABORTED
    assert "Only event version 1.0 is not supported: version='1.1'" in str(
        failed_parse_result.execution.exception
    )

    @contextual_handler(
        parser=example_event_parser,
        context=activate_example_context,
        default_info_factory=ExecutionInfo.from_context,
    )
    def handle_event(event: ExampleEvent) -> ExampleExecutionInfo[str]:
        assert False, "parser aborts so handler is not called"

    with pytest.raises(ExecutionAbortedException) as exc_info:
        handle_event(unsupported_version_example_event_json)

    exc_info.match(re.escape("handle_event input parser rejected input"))


def test_contextual_handler__re_raises_unhandled_exceptions_from_parser_as_ExecutionAbortedException() -> (  # noqa: E501
    None
):
    # Context parser functions should only raise ExecutionAbortedException. This
    # one raises a ValueError without wrapping it in ExecutionAbortedException,
    # which it shouldn't do, because the parser is best placed to record context
    # about the parse error in an ExecutionInfo value.
    def misbehaving_parser(
        raw_json_event: bytes,
    ) -> tuple[ExampleEvent, ExampleContext] | ExecutionInfo:
        raise ValueError("Example unhandled error from parser")

    @contextual_handler(
        parser=misbehaving_parser,
        context=activate_example_context,
        default_info_factory=ExampleExecutionInfo.from_context,
    )
    def handle_event(event: ExampleEvent) -> ExampleExecutionInfo[str]:
        assert False, "parser aborts so handler is not called"

    with pytest.raises(ExecutionAbortedException) as exc_info:
        handle_event(b"not JSON")

    assert exc_info.value.execution_info.execution.status is ExecutionStatus.ABORTED

    assert isinstance(exc_info.value.execution_info.execution.exception, ValueError)
    assert "Example unhandled error from parser" in str(
        exc_info.value.execution_info.execution.exception
    )

    exc_info.match(re.escape("handle_event input context parser threw an exception"))
    exc_info.match(r"input context parser: .*\bmisbehaving_parser")
    exc_info.match(re.escape("input: b'not JSON'"))
    exc_info.match(
        re.escape(
            "Note: context parsers should only throw "
            "ExecutionAbortedException, this parser is misbehaving."
        )
    )


def test_contextual_handler__re_raises_ExecutionAbortedException_raised_from_handler(
    example_event_json: bytes,
) -> None:
    @contextual_handler(
        parser=example_event_parser,
        context=activate_example_context,
        default_info_factory=ExecutionInfo.from_context,
    )
    def handle_event(event: ExampleEvent) -> ExampleExecutionInfo[str]:
        raise ExecutionAbortedException(
            "handle_event threw an ExecutionAbortedException itself",
            execution_info=ExecutionInfo(
                execution=AbortedExecutionResult(
                    reason=AbortedExecutionReason.FUNCTION_ABORTED,
                    exception=RuntimeError("Example error!"),
                )
            ),
        )

    with pytest.raises(ExecutionAbortedException) as exc_info:
        handle_event(example_event_json)

    assert exc_info.value.execution_info.execution.status is ExecutionStatus.ABORTED

    assert isinstance(exc_info.value.execution_info.execution.exception, RuntimeError)
    assert "Example error!" == exc_info.value.execution_info.execution.exception.args[0]

    exc_info.match(re.escape("handle_event threw an ExecutionAbortedException itself"))


def test_contextual_handler__re_raises_exceptions_from_handler_as_ExecutionAbortedException(
    example_event_json: bytes,
) -> None:
    @contextual_handler(
        parser=example_event_parser,
        context=activate_example_context,
        default_info_factory=ExecutionInfo.from_context,
    )
    def handle_event(event: ExampleEvent) -> ExampleExecutionInfo[str]:
        raise RuntimeError("Example error!")

    with pytest.raises(ExecutionAbortedException) as exc_info:
        handle_event(example_event_json)

    assert exc_info.value.execution_info.execution.status is ExecutionStatus.ABORTED

    assert isinstance(exc_info.value.execution_info.execution.exception, RuntimeError)
    assert "Example error!" == exc_info.value.execution_info.execution.exception.args[0]

    exc_info.match(re.escape("handle_event handler threw an exception"))


def test_contextual_handler__has_same_name_as_wrapped_handler_function() -> None:
    @contextual_handler(
        parser=example_event_parser,
        context=activate_example_context,
        default_info_factory=ExecutionInfo.from_context,
    )
    def handle_event(event: ExampleEvent) -> None:
        pass

    assert handle_event.__name__ == "handle_event"



import json

import pytest
from pydantic import ConfigDict, TypeAdapter

from ucam_faas.pydantic import SerialisableException


def test_SerialisableException() -> None:
    # Generate an exception to test with
    with pytest.raises(Exception) as exc_info:
        b"\xff".decode()

    adapter = TypeAdapter[SerialisableException](
        SerialisableException, config=ConfigDict(arbitrary_types_allowed=True)
    )

    # Exceptions are serialised as JSON strings containing the whole traceback
    dumped_json_value = json.loads(adapter.dump_json(exc_info.value))
    assert "Traceback (most recent call last):\n" in dumped_json_value
    assert 'b"\\xff".decode()\n' in dumped_json_value
    assert (
        "UnicodeDecodeError: 'utf-8' codec can't decode byte 0xff in position 0: "
        "invalid start byte\n"
    ) in dumped_json_value

    # Exceptions remain Exceptions when dumped as Python values
    assert adapter.dump_python(exc_info.value) is exc_info.value
    assert adapter.validate_json('"UnicodeDecodeError: blah"') == "UnicodeDecodeError: blah"

    assert adapter.validate_python("UnicodeDecodeError: blah") == "UnicodeDecodeError: blah"
    assert adapter.validate_python(exc_info.value) is exc_info.value



import json
from logging import LogRecord
from pathlib import Path
from typing import Any, Generator

import pytest
from cloudevents import conversion
from cloudevents.abstract import CloudEvent
from faker import Faker
from google.cloud import storage
from pydantic import BaseModel, ConfigDict
from structlog.testing import capture_logs
from ucam_observe import get_structlog_logger  # type: ignore[import-untyped]

from ucam_faas import cloud_event, raw_event, run_ucam_faas
from ucam_faas.cloud_events import CloudEventType, cloud_event_handler

logger = get_structlog_logger(__name__)

script_path = Path(__file__).resolve()
cwd = Path.cwd()
relative_path = script_path.relative_to(cwd)


@pytest.fixture()
def storage_client() -> storage.Client:
    return storage.Client()


@pytest.fixture()
def input_bucket(storage_client: storage.Client) -> Generator[storage.Bucket, None, None]:
    bucket = storage_client.bucket("input-bucket")

    if bucket.exists(client=storage_client):
        bucket.delete(client=storage_client, force=True)

    storage_client.create_bucket(bucket)

    yield bucket

    bucket.delete(client=storage_client, force=True)


@pytest.fixture()
def input_file(faker: Faker, input_bucket: storage.Bucket, valid_cloud_event: CloudEvent) -> str:
    input_file = f"{faker.uuid4()}.event.json"
    blob = input_bucket.blob(input_file)
    blob.upload_from_string(conversion.to_json(valid_cloud_event), content_type="application/json")
    return input_file


@raw_event
def example_raw_event(raw_event: Any) -> None:
    logger.info("example_raw_event", event_content=json.dumps(raw_event))


@cloud_event
def example_cloud_event(event_data: Any) -> None:
    logger.info("example_cloud_event", event_content=json.dumps(event_data))


class ExampleEventData(BaseModel):
    model_config = ConfigDict(extra="allow")
    foo: str


@cloud_event_handler(
    event_data_parser=CloudEventType(
        event_type="ucam_faas_event", event_data_model=ExampleEventData
    )
)
def example_cloud_event_handler(data: ExampleEventData) -> None:
    assert data.foo == "bar"


@cloud_event_handler(
    event_data_parser=CloudEventType(
        event_type="ucam_faas_event", event_data_model=ExampleEventData
    )
)
def example_unsuccessful_cloud_event_handler(data: ExampleEventData) -> None:
    assert data.foo == "bar"
    raise ValueError("Something went wrong")


@pytest.mark.parametrize("target", ["example_raw_event", "example_cloud_event"])
def test_input_file(
    target: str,
    input_file: str,
    input_bucket: storage.Bucket,
    valid_cloud_event: CloudEvent,
) -> None:
    with capture_logs() as cap_logs:
        assert (
            run_ucam_faas(
                target,
                source=relative_path,
                host="no-host",
                port=0,
                debug=False,
                long_running=True,
                input_bucket=input_bucket.name,
                input_file=input_file,
            )
            == 0
        )
    assert len(cap_logs) == 1

    assert "event_content" in cap_logs[0]
    assert json.loads(cap_logs[0]["event_content"]) == valid_cloud_event.get_data()


def test_invalid_input_file(
    input_bucket: storage.Bucket,
) -> None:
    with capture_logs() as cap_logs:
        assert (
            run_ucam_faas(
                "example_raw_event",
                source=relative_path,
                host="no-host",
                port=0,
                debug=False,
                long_running=True,
                input_bucket=input_bucket.name,
                input_file="non_existent_file.event.json",
            )
            == 3
        )
    assert len(cap_logs) == 1
    assert cap_logs[0]["event"] == "Unhandled exception when loading input file"


@pytest.mark.parametrize("target", ["example_raw_event", "example_cloud_event"])
def test_no_input_file(target: str) -> None:
    with capture_logs() as cap_logs:
        assert (
            run_ucam_faas(
                target,
                source=relative_path,
                host="no-host",
                port=0,
                debug=False,
                long_running=True,
            )
            == 0
        )

    assert "event_content" in cap_logs[0]
    assert json.loads(cap_logs[0]["event_content"]) is None


def get_function_result_log(caplog: pytest.LogCaptureFixture) -> tuple[LogRecord, dict[str, Any]]:
    events = ("ucam_faas_function_completed", "ucam_faas_function_aborted")
    record = next(
        (
            r
            for r in caplog.records
            if r.name == "ucam_faas" and isinstance(r.msg, dict) and r.msg["event"] in events
        ),
        None,
    )
    assert (
        record is not None
    ), f"No record was logged in the 'ucam_faas' logger with event {' or '.join(events)}"
    assert isinstance(record.msg, dict)
    return record, record.msg


def test_long_running__executes_cloud_event_handler__and_reports_success(
    input_file: str,
    input_bucket: storage.Bucket,
    valid_cloud_event: CloudEvent,
    caplog: pytest.LogCaptureFixture,
) -> None:
    exit_status = run_ucam_faas(
        "example_cloud_event_handler",
        source=relative_path,
        host="no-host",
        port=0,
        debug=False,
        long_running=True,
        input_bucket=input_bucket.name,
        input_file=input_file,
    )
    assert exit_status == 0

    _, log_msg = get_function_result_log(caplog)

    assert log_msg["event"] == "ucam_faas_function_completed"
    assert log_msg["execution"]["status"] == "COMPLETED"

    assert log_msg["cloudevents"]["event_id"] == valid_cloud_event.get_attributes()["id"]
    assert log_msg["cloudevents"]["event_type"] == "ucam_faas_event"
    assert log_msg["cloudevents"]["event_source"] == "ucam_faas"


def test_long_running__executes_cloud_event_handler__and_reports_failure(
    input_file: str,
    input_bucket: storage.Bucket,
    valid_cloud_event: CloudEvent,
    caplog: pytest.LogCaptureFixture,
) -> None:
    exit_status = run_ucam_faas(
        "example_unsuccessful_cloud_event_handler",
        source=relative_path,
        host="no-host",
        port=0,
        debug=False,
        long_running=True,
        input_bucket=input_bucket.name,
        input_file=input_file,
    )
    assert exit_status == 1

    _, log_msg = get_function_result_log(caplog)

    assert log_msg["event"] == "ucam_faas_function_aborted"
    assert log_msg["execution"]["status"] == "ABORTED"
    assert log_msg["execution"]["reason"] == "ucam_faas.function-aborted"

    assert "Traceback (most recent call last):\n" in log_msg["execution"]["exception"]
    assert "\nValueError: Something went wrong\n" in log_msg["execution"]["exception"]

    assert log_msg["cloudevents"]["event_id"] == valid_cloud_event.get_attributes()["id"]
    assert log_msg["cloudevents"]["event_type"] == "ucam_faas_event"
    assert log_msg["cloudevents"]["event_source"] == "ucam_faas"



"""
Contextual event handler functions receive a primary value as an argument, with
additional context value provided implicitly using a `ContextVar`.

This separation allows a handler function to act primarily on an important subset
of incoming event data, but still access the wider event metadata when needed,
and automatically include metadata in `ExecutionInfo` result values (using the
`contextual_default_factory()` pattern).

Incoming events are separated into the primary value and context value by a
parser function that receives the incoming event value from the parent handler.

The `@contextual_handler(...)` decorator is the interface to this functionality.
"""

from __future__ import annotations

from contextvars import ContextVar
from dataclasses import dataclass
from typing import TYPE_CHECKING, Generic, Protocol, cast, overload

from typing_extensions import Never, TypeVar

from ucam_faas.contexts import ActivateContextVarFn, contextvar_activator
from ucam_faas.exceptions import ExecutionAbortedException, with_notes
from ucam_faas.execution_info import (
    AbortedExecutionReason,
    AbortedExecutionResult,
    CompletedExecutionResult,
    ExecutionInfo,
    ExecutionResult,
)

_InT_con = TypeVar("_InT_con", contravariant=True)
_PrimaryT = TypeVar("_PrimaryT")
_PrimaryT_co = TypeVar("_PrimaryT_co", covariant=True)
_PrimaryT_con = TypeVar("_PrimaryT_con", contravariant=True)
_ContextT = TypeVar("_ContextT")
_ContextT_co = TypeVar("_ContextT_co", covariant=True)

_T_con = TypeVar("_T_con", contravariant=True)
_R_co = TypeVar("_R_co", covariant=True)

_InfoT = TypeVar("_InfoT", bound=ExecutionInfo)
_InfoT_co = TypeVar("_InfoT_co", bound=ExecutionInfo, covariant=True)
_DefaultInfoT_co = TypeVar("_DefaultInfoT_co", bound=ExecutionInfo, covariant=True)
_UsedDefaultInfoT_co = TypeVar(
    "_UsedDefaultInfoT_co", bound=ExecutionInfo, covariant=True, default=Never
)

_NoneT = TypeVar("_NoneT", None, Never, default=Never)
_NoneT_co = TypeVar("_NoneT_co", None, Never, covariant=True, default=Never)


class HandlerFn(Protocol[_T_con, _R_co]):
    """A function wrapped with and called by `@contextual_handler`."""

    def __call__(self, primary_value: _T_con, /) -> _R_co:
        pass

    @property
    def __name__(self) -> str:
        pass


class ContextualHandlerParserFn(Protocol[_InT_con, _PrimaryT_co, _ContextT_co, _InfoT_co]):
    """
    Responsible for parsing a value into a primary result and context metadata.

    The function returns an ExecutionInfo value to abort without further use of
    the input value.
    """

    def __call__(self, input: _InT_con, /) -> tuple[_PrimaryT_co, _ContextT_co] | _InfoT_co:
        pass


class CreateExecutionInfoFn(Protocol[_InfoT_co]):
    def __call__(self, *, execution: ExecutionResult) -> _InfoT_co:
        pass


class ContextualHandlerFn(
    Protocol[_InT_con, _PrimaryT_con, _InfoT_co, _UsedDefaultInfoT_co, _NoneT_co]
):
    def __call__(self, value: _InT_con, /) -> _InfoT_co | _UsedDefaultInfoT_co:
        """
        Validate and do something with a value, and get a description of the
        outcome as an `ExecutionInfo`.
        """

    # Don't define @property at runtime, as they cause subclasses to be read-only
    if TYPE_CHECKING:

        @property
        def __name__(self) -> str:
            pass

        @property
        def handle_parsed_input(self) -> HandlerFn[_PrimaryT_con, _InfoT_co | _NoneT_co]:
            pass


class ContextualHandlerDecoratorFn(Protocol[_InT_con, _PrimaryT, _DefaultInfoT_co]):
    # These overloads are needed to correctly type the return value of
    # ContextualHandlerWrapperFn() according to whether the handler
    # function returns None:
    #   - When the handler returns None, the default ExecutionInfo type is
    #     returned.
    #   - When the handler never returns None, the return is always the
    #     handler's return type.

    @overload
    def __call__(
        self, handler: HandlerFn[_PrimaryT, _InfoT], /
    ) -> ContextualHandlerFn[_InT_con, _PrimaryT, _InfoT]:
        pass

    @overload
    def __call__(
        self, handler: HandlerFn[_PrimaryT, _InfoT | None], /
    ) -> ContextualHandlerFn[_InT_con, _PrimaryT, _InfoT, _DefaultInfoT_co, None]:
        pass

    def __call__(
        self, handler: HandlerFn[_PrimaryT, _InfoT | _NoneT], /
    ) -> ContextualHandlerFn[_InT_con, _PrimaryT, _InfoT, _DefaultInfoT_co, _NoneT]:
        pass


def contextual_handler(
    *,
    parser: ContextualHandlerParserFn[_InT_con, _PrimaryT, _ContextT, ExecutionInfo],
    context: ContextVar[_ContextT] | ActivateContextVarFn[_ContextT],
    default_info_factory: CreateExecutionInfoFn[_DefaultInfoT_co],
) -> ContextualHandlerDecoratorFn[_InT_con, _PrimaryT, _DefaultInfoT_co]:
    """
    Wrap an handler function to execute in a controlled environment.

    The decorated `handler` function is executed:

    - After a `parser` function validates the input value, to extract a primary
      value of interest, and secondary contextual metadata.
    - With the primary value, extracted from an input value by the `parser`
      function as its argument.
    - With the expectation that the `handler` will return an
      `ExecutionInfo` result, or raise an `ExecutionAbortedError` to communicate
      the outcome of the work it has completed using the parsed event.
    - With an ambient context (using a `contextvars.ContextVar`) providing
      access to the secondary contextual metadata parsed from the input value.
      - This context provides access to metadata related to the input that is
        likely less relevant to the `handler` function than the primary value.
      - The ambient context allows components called indirectly within the
        `handler` to access all available information on the input value,
        without the `handler` needing to explicitly pass all the available
        context values. For example, ExecutionInfo types can populate
        their event-related fields automatically in this way, both from the
        immediate context metadata, and from that of (possibly unknown or
        variable) outer contexts wrapping an inner context. This reduces the
        labour involved in reporting detailed ExecutionInfo results, both for
        successful execution and when errors occur.
    - With an `ExecutionInfo` automatically created when the `handler`
      does not return one, or when it throws an exception.
        - This allows for accurate reporting of execution results in the case of
          errors, or when the `handler` opts not to communicate the outcome of
          its work. (Just having completed successfully or not can be enough
          information for simple use-cases.)
    """

    @overload
    def contextual_handler_decorator(
        handler: HandlerFn[_PrimaryT, _InfoT],
    ) -> ContextualHandlerFn[_InT_con, _PrimaryT, _InfoT]:
        pass

    @overload
    def contextual_handler_decorator(
        handler: HandlerFn[_PrimaryT, _InfoT | None],
    ) -> ContextualHandlerFn[_InT_con, _PrimaryT, _InfoT, _DefaultInfoT_co, None]:
        pass

    def contextual_handler_decorator(
        handler: HandlerFn[_PrimaryT, _InfoT | _NoneT],
    ) -> ContextualHandlerFn[_InT_con, _PrimaryT, _InfoT, _DefaultInfoT_co, _NoneT]:
        activate_context: ActivateContextVarFn[_ContextT]
        if isinstance(context, ContextVar):
            activate_context = contextvar_activator(context)
        else:
            activate_context = context

        wrapper = ContextualHandler[
            _InT_con, _PrimaryT, _ContextT, _InfoT, _DefaultInfoT_co, _DefaultInfoT_co, _NoneT
        ](
            handle_parsed_input=handler,
            parse_input=parser,
            activate_context=activate_context,
            create_default_info=default_info_factory,
        )

        # functions_framework relies on the user's function name, so we must preserve it
        assert wrapper.__name__ == handler.__name__  # nosemgrep: bandit.B101

        return wrapper

    return contextual_handler_decorator


@dataclass
class ContextualHandler(
    ContextualHandlerFn[_InT_con, _PrimaryT, _InfoT, _UsedDefaultInfoT_co, _NoneT],
    Generic[
        _InT_con, _PrimaryT, _ContextT, _InfoT, _DefaultInfoT_co, _UsedDefaultInfoT_co, _NoneT
    ],
):
    handle_parsed_input: HandlerFn[_PrimaryT, _InfoT | _NoneT]
    parse_input: ContextualHandlerParserFn[_InT_con, _PrimaryT, _ContextT, ExecutionInfo]
    create_default_info: CreateExecutionInfoFn[_DefaultInfoT_co | _UsedDefaultInfoT_co]
    activate_context: ActivateContextVarFn[_ContextT]

    if TYPE_CHECKING:
        # Handler that does not return None
        @overload
        def __init__(
            self: ContextualHandler[
                # • _UsedDefaultInfoT_co is Never, i.e. call can only return _InfoT
                # • _NoneT is Never, i.e. handle_parsed_input cannot return None
                _InT_con,
                _PrimaryT,
                _ContextT,
                _InfoT,
                _DefaultInfoT_co,
                Never,
                Never,
            ],
            handle_parsed_input: HandlerFn[_PrimaryT, _InfoT],
            parse_input: ContextualHandlerParserFn[_InT_con, _PrimaryT, _ContextT, ExecutionInfo],
            create_default_info: CreateExecutionInfoFn[_DefaultInfoT_co],
            activate_context: ActivateContextVarFn[_ContextT],
        ) -> None:
            pass

        # Handler that does return None
        @overload
        def __init__(
            self: ContextualHandler[
                # • _UsedDefaultInfoT_co is _DefaultInfoT_co, i.e. call can return
                #       _InfoT | _DefaultInfoT_co.
                # • _NoneT is None, i.e. handle_parsed_input can return None.
                _InT_con,
                _PrimaryT,
                _ContextT,
                _InfoT,
                _DefaultInfoT_co,
                _DefaultInfoT_co,
                _NoneT,
            ],
            handle_parsed_input: HandlerFn[_PrimaryT, _InfoT | None],
            parse_input: ContextualHandlerParserFn[_InT_con, _PrimaryT, _ContextT, ExecutionInfo],
            create_default_info: CreateExecutionInfoFn[_DefaultInfoT_co],
            activate_context: ActivateContextVarFn[_ContextT],
        ) -> None:
            pass

        def __init__(
            self,
            handle_parsed_input: HandlerFn[_PrimaryT, _InfoT | None],
            parse_input: ContextualHandlerParserFn[_InT_con, _PrimaryT, _ContextT, ExecutionInfo],
            create_default_info: CreateExecutionInfoFn[_DefaultInfoT_co],
            activate_context: ActivateContextVarFn[_ContextT],
        ) -> None:
            pass

    def __call__(self, input: _InT_con, /) -> _InfoT | _UsedDefaultInfoT_co:
        try:
            parse_result = self.parse_input(input)
            if isinstance(parse_result, ExecutionInfo):
                raise ExecutionAbortedException(
                    f"{self.__name__} input parser rejected input", execution_info=parse_result
                )
        except ExecutionAbortedException:
            raise
        except Exception as e:
            raise with_notes(
                ExecutionAbortedException(
                    f"{self.__name__} input context parser threw an exception",
                    # The context is not activated because the parse function
                    # failed, so we can't use self.create_default_info
                    execution_info=ExecutionInfo.from_context(
                        execution=AbortedExecutionResult(
                            reason=AbortedExecutionReason.FUNCTION_ABORTED, exception=e
                        )
                    ),
                ),
                f"input context parser: {self.parse_input}, input: {input!r}",
                "Note: context parsers should only throw "
                "ExecutionAbortedException, this parser is misbehaving.",
            )

        primary_value, context_value = parse_result

        with self.activate_context(value=context_value):
            execution_info: _InfoT | None
            try:
                execution_info = self.handle_parsed_input(primary_value)
            except ExecutionAbortedException:
                raise
            except Exception as e:
                raise ExecutionAbortedException(
                    f"{self.__name__} handler threw an exception",
                    execution_info=self.create_default_info(
                        execution=AbortedExecutionResult(
                            reason=AbortedExecutionReason.FUNCTION_ABORTED, exception=e
                        )
                    ),
                )

            if execution_info is None:
                # _UsedDefaultInfoT_co IS the same type as _DefaultInfoT_co when
                # execution_info is None. This is enforced by the __init__
                # @overloads.
                return cast(
                    _UsedDefaultInfoT_co,
                    self.create_default_info(execution=CompletedExecutionResult()),
                )
            return execution_info

    # @functions_framework.cloud_event() decorator registers metadata using the
    # __name__ of the decorated function, so we need to preserve the name of the
    # message_handler, so that our user can use the name of their function as
    # its identifier when configuring/deploying their cloud function.
    @property
    def __name__(self) -> str:
        return self.handle_parsed_input.__name__



from __future__ import annotations

import os
import sys
from collections.abc import Callable
from pathlib import Path
from typing import TYPE_CHECKING, Any, Protocol, TypeVar, cast, runtime_checkable

import click
import flask
import flask.typing
import functions_framework
import gunicorn.app.base  # type: ignore[import-untyped]
from cloudevents.abstract import CloudEvent as AbstractCloudEvent
from cloudevents.http import conversion
from cloudevents.http.event import CloudEvent
from google.cloud import storage
from structlog.typing import FilteringBoundLogger
from typing_extensions import ParamSpec
from ucam_observe.gunicorn import logconfig_dict  # type: ignore[import-untyped]
from werkzeug.exceptions import InternalServerError

from ucam_faas._logging import get_structlog_logger
from ucam_faas.cloud_events import (
    CLOUD_EVENT_CONTEXT,
    CloudEventAttributes,
    CloudEventHandlerAttributes,
    CloudEventHandlerContext,
    CloudEventHandlerExecutionInfo,
    CloudEventHandlerExecutorFn,
    CloudEventType,
    cloud_event_handler,
)
from ucam_faas.exceptions import ExecutionAbortedException, UCAMFAASException
from ucam_faas.execution_info import (
    AbortedExecutionReason,
    AbortedExecutionResult,
    CompletedExecutionResult,
    ExecutionInfo,
    ExecutionResult,
    ExecutionStatus,
)
from ucam_faas.gcp_pubsub import (
    PUBSUB_MESSAGE_TYPE,
    PUBSUB_MSG_PUBLISHED_CLOUD_EVENT_TYPE,
    MessagePublishedData,
    PubsubMessage,
)
from ucam_faas.handler_execution import (
    DEFAULT_EXECUTION_RESULT_LOGGER,
    DEFAULT_HTTP_RESPONSE_CREATOR,
    CloudEventFunctionsFrameworkHandlerExecutor,
    CreateExecutionResultHttpResponseFn,
    ExecutionResultLoggerFn,
    HandlerExecutionOptions,
    HttpResponseCreator,
    JsonRepresentation,
    StructlogJsonExecutionResultLogger,
)
from ucam_faas.loading import load_function
from ucam_faas.messages import (
    MESSAGE_CONTEXT,
    MessageHandlerAttributes,
    MessageHandlerContext,
    MessageHandlerExecutionInfo,
    MessageParserFn,
    message_handler,
)

__all__ = (
    "cloud_event_handler",
    "ExecutionAbortedException",
    "ExecutionStatus",
    "AbortedExecutionReason",
    "CompletedExecutionResult",
    "AbortedExecutionResult",
    "ExecutionResult",
    "ExecutionInfo",
    "PUBSUB_MESSAGE_TYPE",
    "PUBSUB_MSG_PUBLISHED_CLOUD_EVENT_TYPE",
    "PubsubMessage",
    "MessagePublishedData",
    "CreateExecutionResultHttpResponseFn",
    "HttpResponseCreator",
    "DEFAULT_HTTP_RESPONSE_CREATOR",
    "ExecutionResultLoggerFn",
    "JsonRepresentation",
    "StructlogJsonExecutionResultLogger",
    "DEFAULT_EXECUTION_RESULT_LOGGER",
    "HandlerExecutionOptions",
    "MESSAGE_CONTEXT",
    "MessageHandlerContext",
    "MessageHandlerAttributes",
    "MessageHandlerExecutionInfo",
    "MessageParserFn",
    "message_handler",
    "message_handler",
    "CLOUD_EVENT_CONTEXT",
    "CloudEventHandlerContext",
    "CloudEventAttributes",
    "CloudEventHandlerAttributes",
    "CloudEventHandlerExecutionInfo",
    "CloudEventType",
    "cloud_event_handler",
    "CloudEventHandlerExecutorFn",
)

if TYPE_CHECKING:
    from _typeshed.wsgi import WSGIApplication

P = ParamSpec("P")
T = TypeVar("T")

# As well as making a logger available this should setup logging before the flask app is created
logger: FilteringBoundLogger = get_structlog_logger(__name__)


class IsRawEvent:
    pass


class IsCloudEvent:
    pass


class HasName(Protocol):
    @property
    def __name__(self) -> str:
        ...


def _common_function_wrapper(function: Callable[P, T]) -> Callable[P, T]:
    def _common_function_wrapper_internal(*args: P.args, **kwargs: P.kwargs) -> T:
        try:
            return function(*args, **kwargs)
        except UCAMFAASException as exception:
            exception_name = exception.__class__.__name__

            logger.warning("function_failed_gracefully", exception_name=exception_name)

            raise InternalServerError(
                description=f"The function raised {exception_name}."
            ) from exception

        except Exception as exception:
            exception_name = exception.__class__.__name__

            logger.exception("function_failed_uncaught_exception", exception_name=exception_name)

            raise

    return _common_function_wrapper_internal


class RawEventHandlerFn(HasName, Protocol):
    def __call__(self, event: bytes, /) -> flask.typing.ResponseReturnValue | None:
        ...


@runtime_checkable
class RegisteredRawEventHandlerFn(HasName, Protocol):
    __ucam_wrapped__: RawEventHandlerFn
    _event: IsRawEvent

    def __call__(self, request: flask.Request, /) -> flask.typing.ResponseReturnValue:
        ...


def raw_event(function: RawEventHandlerFn) -> RegisteredRawEventHandlerFn:
    @_common_function_wrapper
    def _raw_event_internal(request: flask.Request, /) -> flask.typing.ResponseReturnValue:
        return_value = function(request.data)

        if return_value is not None:
            return return_value

        return "", 200

    # Decorators must preserve the wrapped function identity because
    # functions_framework registers metadata against the __name__ of `function`.
    _raw_event_internal.__name__ = function.__name__
    _raw_event_internal = functions_framework.http(_raw_event_internal)

    _raw_event_internal = cast(RegisteredRawEventHandlerFn, _raw_event_internal)
    _raw_event_internal.__ucam_wrapped__ = function
    _raw_event_internal._event = IsRawEvent()

    return _raw_event_internal


class CloudEventHandlerFn(HasName, Protocol):
    def __call__(self, event_data: Any, /) -> None:
        ...


@runtime_checkable
class RegisteredCloudEventHandlerFn(HasName, Protocol):
    __ucam_wrapped__: CloudEventHandlerFn
    _event: IsCloudEvent

    def __call__(self, event: CloudEvent, /) -> None:
        ...


def cloud_event(function: CloudEventHandlerFn) -> RegisteredCloudEventHandlerFn:
    @_common_function_wrapper
    def _cloud_event_internal(event: CloudEvent, /) -> None:
        return function(event.data)

    # Decorators must preserve the wrapped function identity because
    # functions_framework registers metadata against the __name__ of `function`.
    _cloud_event_internal.__name__ = function.__name__
    _cloud_event_internal = functions_framework.cloud_event(_cloud_event_internal)

    _cloud_event_internal = cast(RegisteredCloudEventHandlerFn, _cloud_event_internal)
    _cloud_event_internal.__ucam_wrapped__ = function
    _cloud_event_internal._event = IsCloudEvent()

    return _cloud_event_internal


class FaaSGunicornApplication(gunicorn.app.base.Application):  # type: ignore[misc] # gunicorn is not typed # noqa: E501
    def __init__(self, app: WSGIApplication, host: str, port: int | str) -> None:
        self.host = host
        self.port = port
        self.app = app

        self.options = {
            "bind": "%s:%s" % (host, port),
            "workers": os.environ.get("WORKERS", 2),
            "threads": os.environ.get("THREADS", (os.cpu_count() or 1) * 4),
            "timeout": 0,
            "limit_request_line": 0,
            "logconfig_dict": logconfig_dict,
        }

        super().__init__()

    def load_config(self) -> None:
        for key, value in self.options.items():
            self.cfg.set(key, value)

    def load(self) -> WSGIApplication:
        return self.app


def _initialize_ucam_faas_app(target: str, source: str | Path | None) -> flask.Flask:
    app: flask.Flask = functions_framework.create_app(target, source)  # type: ignore[no-untyped-call] # noqa: E501

    app.logger.info("flask_app_created")

    @app.route("/healthy")
    @app.route("/status")
    def get_status() -> str:
        return "ok"

    return app


def _execute_as_long_running(
    target: str, source: str | Path | None, input_bucket: str | None, input_file: str | Path | None
) -> int:
    """
    Internal function to execute a function as a long-running function (e.g. directly, not served
    as a flask application).

    This method relies on the function registration process having taken place, so must only be
    called after a call to functions_framework.create_app().
    """
    handler = load_function(target, source)

    event = CloudEvent(data=None, attributes={"source": "ucam_faas", "type": "ucam_faas_event"})

    try:
        if input_bucket is not None and input_file is not None:
            client = storage.Client()
            bucket = client.bucket(input_bucket)
            blob = bucket.blob(input_file)
            event = conversion.from_json(blob.download_as_bytes())
    except Exception as exc:
        logger.exception("Unhandled exception when loading input file", exception=exc)
        return 3

    try:
        if isinstance(handler, CloudEventFunctionsFrameworkHandlerExecutor):
            handler = cast(
                CloudEventFunctionsFrameworkHandlerExecutor[
                    Callable[[AbstractCloudEvent], ExecutionInfo]
                ],
                handler,
            )

            execution_info = handler.execute_for_result(event)

            if execution_info.execution.status is ExecutionStatus.ABORTED:
                return 1
            assert execution_info.execution.status is ExecutionStatus.COMPLETED
            return 0

        if not isinstance(handler, (RegisteredRawEventHandlerFn, RegisteredCloudEventHandlerFn)):
            logger.exception("function_handler_not_registered")
            return 2

        handler(event)
    except InternalServerError:
        # This is raised by the `_common_function_wrapper_internal` decorator when
        # `UCAMFAASException` is raised. Logging is already done in the decorator.
        return 1
    except Exception:
        # Logging already done in `_common_function_wrapper_internal`
        return 2
    return 0


def run_ucam_faas(
    target: str,
    source: str | Path | None,
    host: str,
    port: int,
    debug: bool,
    long_running: bool,
    input_bucket: str | None = None,
    input_file: str | Path | None = None,
) -> None | int:  # pragma: no cover
    if long_running:
        return _execute_as_long_running(target, source, input_bucket, input_file)

    # Not long-running, so execute normally
    app = _initialize_ucam_faas_app(target, source)
    if debug:
        app.run(host, port, debug)
    else:
        server = FaaSGunicornApplication(app, host, port)
        server.run()
    return None


@click.command()
@click.option("--target", envvar="FUNCTION_TARGET", type=click.STRING, required=True)
@click.option("--source", envvar="FUNCTION_SOURCE", type=click.Path(), default=None)
@click.option("--host", envvar="HOST", type=click.STRING, default="0.0.0.0")
@click.option("--port", envvar="PORT", type=click.INT, default=8080)
@click.option("--debug", envvar="DEBUG", is_flag=True)
@click.option("--long-running", envvar="LONG_RUNNING", is_flag=True)
@click.option("--input-bucket", envvar="INPUT_BUCKET", type=click.STRING, default=None)
@click.option("--input-file", envvar="INPUT_FILE", type=click.Path(), default=None)
def _cli(
    target: str,
    source: str,
    host: str,
    port: int,
    debug: bool,
    long_running: bool,
    input_bucket: str | None = None,
    input_file: str | Path | None = None,
) -> None:  # pragma: no cover
    return_code = run_ucam_faas(
        target, source, host, port, debug, long_running, input_bucket, input_file
    )

    if long_running:
        sys.exit(return_code)



from __future__ import annotations

from enum import Enum
from typing import Generic, Literal, TypeVar

import pytest
from cloudevents.abstract import CloudEvent
from pydantic import BaseModel, ValidationError
from typing_extensions import Never, assert_type

from ucam_faas.cloud_events import (
    CloudEventAttributes,
    CloudEventHandlerAttributes,
    CloudEventHandlerContext,
    CloudEventHandlerExecutionInfo,
    CloudEventHandlerExecutorFn,
    CloudEventType,
    cloud_event_handler,
    contextual_cloud_event_parser,
)
from ucam_faas.contextual_handlers import ContextualHandlerParserFn
from ucam_faas.execution_info import (
    AbortedExecutionReason,
    AbortedExecutionResult,
    ExecutionInfo,
    ExecutionResult,
    ExecutionStatus,
)
from ucam_faas.tests.fixtures import make_cloud_event

_T = TypeVar("_T")


class DataExecutionInfo(CloudEventHandlerExecutionInfo, Generic[_T]):
    data: _T


def make_example_cloud_event(data: object) -> CloudEvent:
    return make_cloud_event(
        type="com.example.sampletype1", source="https://example.com/event-producer", data=data
    )


@pytest.fixture
def cloud_event() -> CloudEvent:
    return make_example_cloud_event(data=None)


class SpecVersion(Enum):
    V1_0 = "1.0"


def parse_specversion(cloud_event: CloudEvent) -> SpecVersion:
    return SpecVersion(cloud_event.get_attributes()["specversion"])


def test_cloud_event_handler__may_return_none(cloud_event: CloudEvent) -> None:
    @cloud_event_handler(event_data_parser=parse_specversion)
    def handler(spec_version: SpecVersion) -> None:
        return None

    assert_type(
        handler,
        CloudEventHandlerExecutorFn[SpecVersion, Never, CloudEventHandlerExecutionInfo, None],
    )

    execution_info = handler.execute(cloud_event)

    assert_type(execution_info, CloudEventHandlerExecutionInfo)
    assert execution_info.execution.status is ExecutionStatus.COMPLETED


def test_cloud_event_handler__returns_specified_default_ExecutionInfo_type_when_handler_returns_none(  # noqa: E501
    cloud_event: CloudEvent,
) -> None:
    def create_default_execution_info(
        *, execution: ExecutionResult
    ) -> DataExecutionInfo[Literal["foo"]]:
        return DataExecutionInfo[Literal["foo"]](data="foo", execution=execution)

    @cloud_event_handler(
        event_data_parser=parse_specversion, create_default_info=create_default_execution_info
    )
    def handler(spec_version: SpecVersion) -> None:
        return None

    assert_type(
        handler,
        CloudEventHandlerExecutorFn[SpecVersion, Never, DataExecutionInfo[Literal["foo"]], None],
    )

    execution_info = handler.execute(cloud_event)

    assert_type(execution_info, DataExecutionInfo[Literal["foo"]])
    assert execution_info.execution.status is ExecutionStatus.COMPLETED
    assert execution_info.data == "foo"


def test_cloud_event_handler__returns_only_handler_type_when_handler_does_not_return_none(
    cloud_event: CloudEvent,
) -> None:
    @cloud_event_handler(event_data_parser=parse_specversion)
    def handler(specversion: SpecVersion) -> DataExecutionInfo[SpecVersion]:
        return DataExecutionInfo(data=specversion)

    assert_type(
        handler,
        CloudEventHandlerExecutorFn[SpecVersion, DataExecutionInfo[SpecVersion], Never, Never],
    )

    execution_info = handler.execute(cloud_event)

    assert_type(execution_info, DataExecutionInfo[SpecVersion])
    assert execution_info.execution.status == ExecutionStatus.COMPLETED
    assert execution_info.data is SpecVersion.V1_0


class TestCloudEventType:
    class ExampleEventData(BaseModel):
        greeting: str

    @pytest.fixture
    def valid_event_data(self) -> TestCloudEventType.ExampleEventData:
        return TestCloudEventType.ExampleEventData(greeting="Hello World!")

    @pytest.fixture
    def valid_event(self, valid_event_data: TestCloudEventType.ExampleEventData) -> CloudEvent:
        return make_cloud_event(
            type="com.example.sampletype1",
            source="https://example.com/event-producer",
            data=valid_event_data.model_dump(mode="json"),
        )

    @pytest.fixture
    def invalid_event_wrong_type(
        self, valid_event_data: TestCloudEventType.ExampleEventData
    ) -> CloudEvent:
        return make_cloud_event(
            type="com.example.sampletype2",
            source="https://example.com/event-producer",
            data=valid_event_data.model_dump(mode="json"),
        )

    @pytest.fixture
    def invalid_event_invalid_data(self) -> CloudEvent:
        return make_cloud_event(
            type="com.example.sampletype1",
            source="https://example.com/event-producer",
            data={"greeting": 42},  # invalid because 42 should be a string
        )

    @pytest.fixture
    def cloud_event_type(self) -> CloudEventType[TestCloudEventType.ExampleEventData]:
        return CloudEventType(
            event_type="com.example.sampletype1",
            event_data_model=TestCloudEventType.ExampleEventData,
        )

    def test_parses_valid_cloudevent(
        self,
        cloud_event_type: CloudEventType[TestCloudEventType.ExampleEventData],
        valid_event: CloudEvent,
        valid_event_data: TestCloudEventType.ExampleEventData,
    ) -> None:
        assert cloud_event_type(valid_event) == valid_event_data

    def test_rejects_event_with_wrong_type(
        self,
        cloud_event_type: CloudEventType[TestCloudEventType.ExampleEventData],
        invalid_event_wrong_type: CloudEvent,
    ) -> None:
        # contextual_cloud_event_parser() is needed to provide the context to create
        # CloudEventHandlerExecutionInfo when the parser rejects the event.
        parse_result = contextual_cloud_event_parser(cloud_event_type)(invalid_event_wrong_type)

        assert isinstance(parse_result, CloudEventHandlerExecutionInfo)
        assert parse_result.execution.status is ExecutionStatus.ABORTED
        assert parse_result.execution.reason is AbortedExecutionReason.CLOUD_EVENT_TYPE_INCORRECT
        assert isinstance(parse_result.execution.exception, ValueError)
        assert "Received CloudEvent with incorrect type: 'com.example.sampletype2'" in str(
            parse_result.execution.exception
        )
        assert parse_result.cloudevents == CloudEventAttributes.from_cloud_event(
            invalid_event_wrong_type
        )
        assert parse_result.cloud_event_handler.input_event_accepted is False

    def test_rejects_event_with_invalid_data(
        self,
        cloud_event_type: CloudEventType[TestCloudEventType.ExampleEventData],
        invalid_event_invalid_data: CloudEvent,
    ) -> None:
        # contextual_cloud_event_parser() is needed to provide the context to create
        # CloudEventHandlerExecutionInfo when the parser rejects the event.
        parse_result = contextual_cloud_event_parser(cloud_event_type)(invalid_event_invalid_data)

        assert isinstance(parse_result, CloudEventHandlerExecutionInfo)
        assert parse_result.execution.status is ExecutionStatus.ABORTED
        assert parse_result.execution.reason is AbortedExecutionReason.SERVICE_DATA_INVALID

        assert isinstance(parse_result.execution.exception, ValidationError)
        exc_info = pytest.ExceptionInfo.from_exception(parse_result.execution.exception)
        exc_info.match(r"validation error for ExampleEventData")
        exc_info.match(r"greeting\s+Input should be a valid string")

        assert parse_result.cloudevents == CloudEventAttributes.from_cloud_event(
            invalid_event_invalid_data
        )
        assert parse_result.cloud_event_handler.input_event_accepted is False


def test_contextual_cloud_event_parser() -> None:
    @contextual_cloud_event_parser
    def parse_cloud_event(cloud_event: CloudEvent, /) -> Literal["Example"] | ExecutionInfo:
        data = cloud_event.get_data()
        if data == "Example":
            return "Example"
        return CloudEventHandlerExecutionInfo(
            execution=AbortedExecutionResult(reason=AbortedExecutionReason.SERVICE_DATA_INVALID)
        )

    assert_type(
        parse_cloud_event,
        ContextualHandlerParserFn[
            CloudEvent, Literal["Example"], CloudEventHandlerContext, ExecutionInfo
        ],
    )

    # Successfully-parsed data becomes the primary value, with the original
    # CloudEvent in the context data.
    cloud_event = make_example_cloud_event(data="Example")
    assert parse_cloud_event(cloud_event) == (
        "Example",
        CloudEventHandlerContext(
            cloud_event=cloud_event,
            execution_id=None,
            handler=CloudEventHandlerAttributes(input_event_accepted=True),
        ),
    )

    # Unsuccessful parse result ExecutionInfo is returned as-is.
    bad_cloud_event = make_example_cloud_event(data="Invalid")
    assert parse_cloud_event(bad_cloud_event) == CloudEventHandlerExecutionInfo(
        execution=AbortedExecutionResult(reason=AbortedExecutionReason.SERVICE_DATA_INVALID),
        execution_id=None,
        cloudevents=CloudEventAttributes.from_cloud_event(bad_cloud_event),
        cloud_event_handler=CloudEventHandlerAttributes(input_event_accepted=False),
    )



import re
import sys

import pytest

from ucam_faas.exceptions import ExecutionAbortedException, with_notes
from ucam_faas.execution_info import AbortedExecutionResult, ExecutionInfo


@pytest.mark.parametrize(
    "args, repr_template",
    [
        pytest.param(
            ("The magic smoke escaped",),
            "ExecutionAbortedException('The magic smoke escaped', execution_info={})",
            id="1 arg",
        ),
        pytest.param(
            ("The magic smoke escaped", 42),
            "ExecutionAbortedException('The magic smoke escaped', 42, execution_info={})",
            id="2 args",
        ),
    ],
)
def test_ExecutionAbortedException(args: tuple[object, ...], repr_template: str) -> None:
    execution_info = ExecutionInfo(execution=AbortedExecutionResult(reason="magicsmoke"), foo=123)
    err = ExecutionAbortedException(*args, execution_info=execution_info)

    assert err.execution_info is execution_info
    assert err.args == args
    assert repr(err) == repr_template.format(repr(execution_info))


@pytest.mark.skipif(condition=sys.version_info < (3, 11), reason="Notes are not rendered < py311")
def test_with_notes__results_in_notes_displaying_from_311() -> None:
    with pytest.raises(
        ValueError,
        match=re.compile(r"boom.*Useful context.*More info", re.MULTILINE | re.DOTALL),
    ):
        raise with_notes(ValueError("boom"), "Useful context", "More info")


@pytest.mark.skipif(condition=sys.version_info >= (3, 11), reason="Notes are rendered >= py311")
def test_with_notes__does_not_fail_pre_311() -> None:
    # we don't expect notes to render, just that they don't break
    with pytest.raises(ValueError, match=r"boom") as exc_info:
        raise with_notes(ValueError("boom"), "Useful context", "More info")

    assert "Useful context" in getattr(exc_info.value, "__notes__")
    assert "More info" in getattr(exc_info.value, "__notes__")



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: example.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\rexample.proto"M\n\rExampleRecord\x12\r\n\x05query\x18\x01 \x01(\t\x12\x13\n\x0bpage_number\x18\x02 \x01(\x05\x12\x18\n\x10results_per_page\x18\x03 \x01(\x05\x62\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "example_pb2", globals())
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _EXAMPLERECORD._serialized_start = 17
    _EXAMPLERECORD._serialized_end = 94
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

import warnings
from typing import Generator, cast

from structlog.typing import FilteringBoundLogger
from ucam_observe import get_structlog_logger as _get_structlog_logger  # type: ignore


def get_structlog_logger(name: str = "ucam_faas") -> FilteringBoundLogger:
    """Get a logger that emits a JSON object for each log call."""
    return cast(FilteringBoundLogger, _get_structlog_logger(name))


# As well as making a logger available this should setup logging before the flask app is created
logger: FilteringBoundLogger = get_structlog_logger(__name__)


class FlattenedJsonPropertyClashWarning(UserWarning):
    pass


def flatten_json(json_value: object) -> dict[str, object]:
    """
    Flatten a JSON tree into a single wide object containing scalar leaf values.

    Each value's property name is its path in the input object, joined with ".".

    Examples
    --------
    >>> wide_obj = flatten_json({
    ...     "ok": True,
    ...     "person": {"name": "Joe", "age": 100},
    ...     "pets": [
    ...         {"species": "cat", "colour": "orange"},
    ...         {"species": "dog", "size": "tiny"},
    ...     ],
    ...     "empty": {},
    ... })

    >>> import json
    >>> print(json.dumps(wide_obj, indent=2))
    {
      "ok": true,
      "person.name": "Joe",
      "person.age": 100,
      "pets.0.species": "cat",
      "pets.0.colour": "orange",
      "pets.1.species": "dog",
      "pets.1.size": "tiny"
    }

    # The flat representation is not lossless, if two values have the same
    # property path (due to one or both containing "." in their own keys) then
    # the last value will override earlier values.
    >>> import pytest
    >>> with pytest.warns(FlattenedJsonPropertyClashWarning,
    ...                   match=r'value was lost when flattening a JSON value'):
    ...     flatten_json({"foo": {"bar": 42}, "foo.bar": 100})
    {'foo.bar': 100}
    """
    if not isinstance(json_value, dict):
        json_value = {"message": json_value}

    items = [
        (".".join(map(str, path)), value)
        for (path, value) in iter_json_tree_leaves((), json_value)
    ]
    flat_props = {path: value for (path, value) in items}
    if len(items) > len(flat_props):
        warnings.warn(
            "At least one value was lost when flattening a JSON value, because "
            "two or more values have the same dot-separated property path.",
            category=FlattenedJsonPropertyClashWarning,
            stacklevel=2,
        )
    return flat_props


def iter_json_tree_leaves(
    path: tuple[str, ...], inner_node: dict[str, object] | list[object]
) -> Generator[tuple[tuple[int | str, ...], object]]:
    """
    Walk the leaf nodes (scalar values) of a JSON tree.

    This is a generator that yields tuple pairs containing:
    1. The path of the leaf node from the root to the value, as a tuple of str | int values
    2. The value
    """
    items = inner_node.items() if isinstance(inner_node, dict) else enumerate(inner_node)
    for name, value in items:
        sub_path = (*path, str(name))
        if isinstance(value, (dict, list)):
            yield from iter_json_tree_leaves(sub_path, value)
        else:
            yield (sub_path, value)



from __future__ import annotations

import importlib.util
import os
import sys
from pathlib import Path
from types import FunctionType, ModuleType

from ucam_faas.exceptions import UCAMFAASCouldNotLoadTarget


def load_function(target: str, source: str | Path | None) -> FunctionType:
    # If this raises an exception it is propogated up
    module = _load_source_module(source or "./main.py")

    if not hasattr(module, target):
        raise UCAMFAASCouldNotLoadTarget(
            f"Source file {source} did not have expected target function {target}"
        )
    func = getattr(module, target)
    if not isinstance(func, FunctionType):
        raise UCAMFAASCouldNotLoadTarget(
            f"Target {target} in source file {source} does not appear to be a valid function"
        )
    return func


def _load_source_module(source: str | Path) -> ModuleType:
    realpath = os.path.realpath(source)
    dir, filename = os.path.split(realpath)
    name, _ = os.path.splitext(filename)
    spec = importlib.util.spec_from_file_location(name, realpath, submodule_search_locations=[dir])
    if spec and spec.loader:
        source_module = importlib.util.module_from_spec(spec)
        sys.path.append(dir)
        sys.modules[name] = source_module
        spec.loader.exec_module(source_module)
        return source_module
    raise UCAMFAASCouldNotLoadTarget(
        f"Source file {source} could not be loaded as a python module"
    )



from __future__ import annotations

import contextlib
import os
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Generator, Protocol

from flask.testing import FlaskClient

from . import _initialize_ucam_faas_app

if TYPE_CHECKING:
    from typing_extensions import Never


class CreateEventAppClientFn(Protocol):
    def __call__(self, target: str, source: str | Path | None = None) -> FlaskClient:
        ...


try:
    if TYPE_CHECKING:
        from cloudevents.pydantic.v2 import CloudEvent
    else:
        from cloudevents.pydantic import CloudEvent
    from polyfactory.factories.pydantic_factory import ModelFactory
    from pytest import fixture

    @fixture
    def event_app_test_client_factory(cleanup_functions_framework: None) -> CreateEventAppClientFn:
        # Although we use the cleanup_functions_framework fixture, tests that
        # call event_app_test_client_factory() multiple times will still be able
        # to see global state changes made by functions_framework in previous
        # calls to event_app_test_client_factory() within the same test. We
        # could make _event_app_client() a context manager to clean each call
        # rather than at the per-test level.

        def _event_app_client(target: str, source: str | Path | None = None) -> FlaskClient:
            test_app = _initialize_ucam_faas_app(target, source)
            return test_app.test_client()

        return _event_app_client

    class CloudEventFactory(ModelFactory[CloudEvent]):
        __model__ = CloudEvent

        specversion = "1.0"

except ImportError as e:
    _import_error = e

    @fixture
    def event_app_test_client_factory() -> Never:  # type: ignore[misc]
        raise RuntimeError(
            f"Fixture {event_app_test_client_factory.__name__} is not available "
            f"because ucam_faas is not installed with the 'testing' extra]"
        ) from _import_error


@fixture(scope="function", name="cleanup_functions_framework")
def cleanup_functions_framework_fixture() -> Generator[None]:
    with cleanup_functions_framework():
        yield


@contextlib.contextmanager
def cleanup_functions_framework() -> Generator[None]:
    """
    A context manager that resets changes made to global Python interpreter
    state by `functions_framework` within the managed context.

    `functions_framework` changes environment variables and changes `sys.path`.
    These changes persist between subsequent `functions_framework.create_app()`
    calls, resulting in subsequent calls being affected by state changed by
    previous calls. (Specifically, envars that remain set cause subsequent
    functions to be registered with the wrong function type.)

    Wrapping calls to `functions_framework` within this context manager will
    prevent global changes affecting subsequent calls to `functions_framework`.
    """
    mutated_envars = ["FUNCTION_TARGET", "FUNCTION_SIGNATURE_TYPE", "FUNCTION_TRIGGER_TYPE"]
    initial_envars = {e: os.environ.get(e) for e in mutated_envars}

    # functions_framework appends to sys.path without checking if an entry
    # already exists. This results in the path increasing in size with each
    # call. To limit growth, we remove duplicate entries from sys.path.
    initial_syspath = list(sys.path)

    yield

    for name, initial_value in initial_envars.items():
        if initial_value is None:
            os.environ.pop(name, None)
        else:
            os.environ[name] = initial_value

    sys.path[:] = initial_syspath
